"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const axios_1 = __importDefault(require("axios"));
const operators_1 = require("rxjs/operators");
const types_1 = require("../../types");
const debug = require('debug')('exchange-aggregator:kyber');
const getCurrenciesHttpUrl = (network) => {
    switch (network) {
        case types_1.Network.MAINNET:
            return 'https://api.kyber.network/currencies';
        default:
            throw new Error('Kyber only supports the MAINNET network.');
    }
};
const getRateHttpUrl = (network, type, currency, interval) => {
    const prefix = `https://api.kyber.network/${type}_rate`;
    const splits = R.splitEvery(5, interval);
    const suffix = splits
        .map(split => {
        const quantities = split
            .map(qty => {
            return `qty=${qty}`;
        })
            .join('&');
        return `id=${currency}&${quantities}`;
    })
        .join('&');
    switch (network) {
        case types_1.Network.MAINNET:
            return `${prefix}?${suffix}`;
        default:
            throw new Error('Kyber only supports the MAINNET network.');
    }
};
exports.getObservableKyberOrders = ({ base, quote, network }) => {
    if (quote !== 'ETH') {
        throw new Error('Kyber only support ETH as a quote token.');
    }
    const currencies$ = Rx.defer(() => {
        const url = getCurrenciesHttpUrl(network);
        return Rx.from(axios_1.default.get(url).then(result => result.data));
    }).pipe(operators_1.map(result => result.data));
    const formatResponse = (type, response) => {
        if (response.error) {
            throw new Error(`Error trying to fetch Kyber ${type} rates.`);
        }
        const volumeKey = type === 'buy' ? 'dst_qty' : 'src_qty';
        const priceKey = type === 'buy' ? 'src_qty' : 'dst_qty';
        const groups = response.data.map(current => {
            return current[volumeKey].map((volume, index) => ({
                volume,
                price: current[priceKey][index],
            }));
        });
        return [].concat(...groups);
    };
    const pollRate = (currencies) => {
        const currency = currencies.find(R.propEq('symbol', base));
        if (!currency) {
            throw new Error(`The ${base} token is not supported.`);
        }
        const interval = [1, 10, 20, 30, 40, 50, 75, 100, 125, 150];
        return Rx.interval(5000).pipe(operators_1.switchMap(() => {
            debug(`Loading snapshots for market %s-%s.`, base, quote);
            const buyUrl = getRateHttpUrl(network, 'buy', currency.id, interval);
            const sellUrl = getRateHttpUrl(network, 'sell', currency.id, interval);
            const sellRequest = axios_1.default
                .get(sellUrl)
                .then(result => result.data);
            const buyRequest = axios_1.default
                .get(buyUrl)
                .then(result => result.data);
            return Rx.from(Promise.all([buyRequest, sellRequest]));
        }), operators_1.distinctUntilChanged(), operators_1.map(([buyResponse, sellResponse]) => ({
            base,
            quote,
            asks: formatResponse('buy', buyResponse),
            bids: formatResponse('sell', sellResponse),
        })), operators_1.retryWhen(error => {
            debug('Kyber failed with error.');
            return error.pipe(operators_1.delay(10000));
        }));
    };
    const rates$ = currencies$.pipe(operators_1.switchMap(pollRate));
    return rates$;
};
exports.standardizeStream = (stream$) => {
    return stream$.pipe(operators_1.map(orderbook => ({
        event: types_1.StandardizedMessageType.SNAPSHOT,
        exchange: types_1.Exchange.KYBER,
        base: orderbook.base,
        quote: orderbook.quote,
        asks: orderbook.asks,
        bids: orderbook.bids,
    })));
};
