"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const axios_1 = __importDefault(require("axios"));
const operators_1 = require("rxjs/operators");
const webSocket_1 = require("rxjs/webSocket");
const isomorphic_ws_1 = __importDefault(require("isomorphic-ws"));
const types_1 = require("@radarrelay/types");
const types_2 = require("../../types");
const debug = require('debug')('exchange-aggregator:radar-relay');
const subscribeMessage = (base, quote) => {
    const baseToken = cleanToken(base);
    const quoteToken = cleanToken(quote);
    const message = {
        type: types_1.WebsocketRequestType.SUBSCRIBE,
        topic: types_1.WebsocketRequestTopic.BOOK,
        market: `${baseToken}-${quoteToken}`,
    };
    return message;
};
const standardizeOrder = (order) => ({
    price: parseFloat(order.price),
    volume: parseFloat(order.remainingBaseTokenAmount),
    metadata: order,
});
const standardizeNewOrderEvent = (event) => ({
    event: types_2.StandardizedMessageType.ADD,
    exchange: types_2.Exchange.RADAR_RELAY,
    type: event.order.type,
    id: event.order.orderHash,
    order: standardizeOrder(event.order),
});
const standardizeFillOrderEvent = (event) => ({
    event: types_2.StandardizedMessageType.FILL,
    exchange: types_2.Exchange.RADAR_RELAY,
    type: event.order.type,
    id: event.order.orderHash,
    order: standardizeOrder(event.order),
});
const standardizeCancelOrderEvent = (event) => ({
    event: types_2.StandardizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    type: event.orderType,
    id: event.orderHash,
});
const standardizeRemoveOrderEvent = (event) => ({
    event: types_2.StandardizedMessageType.REMOVE,
    exchange: types_2.Exchange.RADAR_RELAY,
    type: event.orderType,
    id: event.orderHash,
});
const standardizeSnapshotEvent = (book) => ({
    event: types_2.StandardizedMessageType.SNAPSHOT,
    exchange: types_2.Exchange.RADAR_RELAY,
    asks: book.asks.map(standardizeOrder),
    bids: book.bids.map(standardizeOrder),
});
exports.isSnapshotEvent = R.allPass([R.has('asks'), R.has('bids')]);
exports.isOrderEvent = R.allPass([R.has('event'), R.has('action')]);
exports.isNewOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'NEW'),
]);
exports.isFillOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'FILL'),
]);
exports.isCancelOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'CANCEL'),
]);
exports.isRemoveOrderEvent = R.allPass([
    exports.isOrderEvent,
    R.propEq('action', 'REMOVE'),
]);
const getWebsocketUrl = (network) => {
    switch (network) {
        case types_2.Network.KOVAN:
            return 'wss://ws.kovan.radarrelay.com/v2';
        case types_2.Network.MAINNET:
            return 'wss://ws.radarrelay.com/v2';
        default:
            throw new Error('Invalid network.');
    }
};
const cleanToken = (token) => {
    switch (token) {
        case 'ETH':
            return 'WETH';
        default:
            return token;
    }
};
const getHttpUrl = (base, quote, network) => {
    const baseToken = cleanToken(base);
    const quoteToken = cleanToken(quote);
    switch (network) {
        case types_2.Network.KOVAN:
            return `https://api.kovan.radarrelay.com/v2/markets/${baseToken}-${quoteToken}/book`;
        case types_2.Network.MAINNET:
            return `https://api.radarrelay.com/v2/markets/${baseToken}-${quoteToken}/book`;
        default:
            throw new Error('Invalid network.');
    }
};
exports.getObservableRadarRelayOrders = ({ base, quote, network, }) => {
    const open$ = new Rx.Subject();
    const close$ = new Rx.Subject();
    const ws$ = webSocket_1.webSocket({
        WebSocketCtor: isomorphic_ws_1.default,
        closeObserver: close$,
        openObserver: open$,
        url: getWebsocketUrl(network),
    });
    open$.subscribe(() => {
        debug('Opening connection.');
        ws$.next(subscribeMessage(base, quote));
    });
    close$.subscribe(() => {
        debug('Closing connection.');
    });
    const socket$ = ws$.pipe(operators_1.retryWhen(errors => {
        debug('Connection error. Retrying after a delay.', errors);
        return errors.pipe(operators_1.delay(1000));
    }));
    // Send a ping signal after a certain quiet period (no message
    // received within 5 seconds).
    const ping$ = socket$.pipe(operators_1.skipUntil(open$), operators_1.takeUntil(close$), operators_1.switchMap(() => Rx.timer(5000)));
    ping$.subscribe(() => {
        debug('Sending ping signal.');
        ws$.next('ping');
    });
    const snapshot$ = socket$.pipe(operators_1.filter(R.allPass([
        R.propEq('type', types_1.WebsocketRequestType.SUBSCRIBE),
        R.propEq('topic', types_1.WebsocketRequestTopic.BOOK),
    ])), operators_1.switchMap(() => {
        debug(`Loading snapshot for market ${base}-${quote}.`);
        const url = getHttpUrl(base, quote, network);
        return Rx.from(axios_1.default.get(url).then(result => result.data));
    }), operators_1.tap(data => {
        debug('Receiving snapshot (%s bids and %s asks).', data.bids.length, data.asks.length);
    }));
    const messages$ = socket$.pipe(operators_1.filter(exports.isOrderEvent), operators_1.tap(value => {
        debug('Receiving %s message.', value.action);
    }));
    // TODO: Make it so, that every time we are fetching a new snapshot (every
    // time we (re-)open the connection), the websocket messages are buffered
    // until the snapshot is emitted.
    return Rx.merge(snapshot$, messages$).pipe(operators_1.share());
};
exports.standardizeStream = (stream$) => {
    const usingEvent = (fn) => R.compose(fn, R.prop('event'));
    return stream$.pipe(operators_1.map(R.cond([
        [exports.isNewOrderEvent, usingEvent(standardizeNewOrderEvent)],
        [exports.isFillOrderEvent, usingEvent(standardizeFillOrderEvent)],
        [exports.isCancelOrderEvent, usingEvent(standardizeCancelOrderEvent)],
        [exports.isRemoveOrderEvent, usingEvent(standardizeRemoveOrderEvent)],
        [exports.isSnapshotEvent, standardizeSnapshotEvent],
    ])));
};
